# -*- coding: utf-8 -*-
"""
Manage Blueprints.
"""

import os
import re
import sys
import clc
import uuid
import json
import time
from lxml import objectify, etree

import bpformation


# TODO vCur:


#
# TODO vNext:
#  o Create (via wizard)
#  o Blueprint package parameters - verify with /blueprints/designer/TaskParameters and alert on error
#  o Dependencies tree


class Blueprint():

	visibility_stoi = { 'public': 1, 'private': 2, 'shared': 3}
	limited_printable_chars = "0123456789abcdefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXYZ!\"#$%&'()*+,-./:;<=>?@[]^_`{|}~  "


	@staticmethod
	def Delete(ids):
		for id in ids:
			# Lookup package uuid given id
			r = bpformation.web.CallScrape("GET","/blueprints/browser/details/%s" % id)
			try:
				name = re.search('<h1 id="body-title" class="left">\s*(.+?)\s*<small>',r.text,re.DOTALL).group(1)
				bp_uuid = re.search("/Blueprints/Designer/MetaData/(.+?)\"",r.text).group(1)
			except:
				bpformation.output.Status('ERROR',3,"Unable to location Blueprint %s (status code %s)" % (id,r.status_code))
				continue
				

			# Exec delete against uuid
			r = bpformation.web.CallScrape("POST","/blueprints/browser/delete", payload={"uuid": bp_uuid, })
			if r.status_code<400 and r.status_code>=200:
				bpformation.output.Status('SUCCESS',3,"%s deleted (id %s)" % (name,id))
			else:
				bpformation.output.Status('ERROR',3,"%s deletion error (status code %s)" % (id,r.status_code))


	@staticmethod
	def _ParseExportTaskDeployPackage(root):
		package_obj = {"type": "package", "id": root.get("ID"), "uuid": root.get("UUID"), "name": root.get("Title"), 'properties': {} }
		if root.get("Server"):  package_obj['server'] = root.get("Server")
		for prop in root.iter("Property"):
			package_obj['properties'][prop.get("Name")] = prop.get("Value")

		if not len(package_obj['properties'].keys()):  del(package_obj['properties'])
		return(package_obj)


	@staticmethod
	def _ParseExportTaskAddIPAddress(root):
		ip_obj = {"type": "add_ip", "id": root.get("ID"), "uuid": root.get("UUID") }
		return(ip_obj)


	@staticmethod
	def _ParseExportTaskAddMappedIPAddress(root):
		ip_obj = {"type": "add_nat_ip", "id": root.get("ID"), "uuid": root.get("UUID"),
		          "ingress_ports": root.get("FirewallOptions").split(", ") }
		return(ip_obj)


	@staticmethod
	def _ParseExportTaskReboot(root):
		ip_obj = {"type": "reboot", "id": root.get("ID"), "uuid": root.get("UUID") }
		return(ip_obj)


	@staticmethod
	def _ParseExportTaskAddDisk(root):
		disk_obj = {"type": "disk", "id": root.get("ID"), "uuid": root.get("UUID")}
		for prop in root[0]:
			if prop.get("Value")=="True":  disk_obj[prop.get("Name").lower()] = True
			elif prop.get("Value")=="False":  disk_obj[prop.get("Name").lower()] = False
			else:  disk_obj[prop.get("Name").lower()] = prop.get("Value")

		return(disk_obj)


	@staticmethod
	def _ParseExportTaskBuildServer(root):
		# Server foundation data
		server = { 'type': "server", 'name': root.get("Alias"), 'uuid': root.get("UUID"), 'description': root.get("Description"),
				   'id': root.get("ID"), 'template': root.get("Template"), 'cpu': root.get("CpuCount"), 
				   'ram': root.get("MemoryGB"), 'tasks': [] }

		for o in root:
			if o.tag=='DeployPackage':  server['tasks'].append(Blueprint._ParseExportTaskDeployPackage(o))
			elif o.tag=='AddDisk':  server['tasks'].append(Blueprint._ParseExportTaskAddDisk(o))
			elif o.tag=='AddIPAddress':  server['tasks'].append(Blueprint._ParseExportTaskAddIPAddress(o))
			elif o.tag=='AddMappedIPAddress':  server['tasks'].append(Blueprint._ParseExportTaskAddMappedIPAddress(o))
			elif o.tag=='Reboot':  server['tasks'].append(Blueprint._ParseExportTaskReboot(o))
			elif o.tag=='Properties':  continue
			else:  bpformation.output.Status('ERROR',3,"Error unknown server tag '%s'" % o.tag)

		return(server)


	@staticmethod
	def _ExportProcessRoot(root):
		tasks = []
		if root.tag=="BuildServer":  tasks.append(Blueprint._ParseExportTaskBuildServer(root))
		elif root.tag=="DeployPackage":  tasks.append(Blueprint._ParseExportTaskDeployPackage(root))
		elif root.tag=="Blueprint":  
			for o in root:
				try:
					if o.tag != "Blueprint":  tasks.append(Blueprint._ExportProcessRoot(o)[0])
				except:
					# Catch only our exceptions
					pass
		else:  
			# Not currently supported: Reboot outside of build server
			print "Unknown: %s" % root.tag
			raise(Exception("Unknown tag %s" % root.tag))

		return(tasks)


	@staticmethod
	def Export(id,file=None):
		# Silence status output if writing to stdout
		if file == "-":  bpformation.args.args.quiet = 999

		bp = {'metadata': {}, 'tasks': [], 'execute': {} }

		# Blueprint metadata 
		r = bpformation.web.CallScrape("GET","/blueprints/browser/details/%s" % id)
		if r.status_code<200 or r.status_code>=300:
			bpformation.output.Status('ERROR',3,"Error retrieving data (http response %s)" % r.status_code)
			raise(bpformation.BPFormationFatalExeption("Fatal Error"))
		bp['metadata'] = {
				'name': re.search('<h1 id="body-title" class="left">\s*(.+?)\s*<small>',r.text,re.DOTALL).group(1),
				'owner': re.search('<small>by (.+?)</small>',r.text).group(1),
				'version': re.search('<dt>version</dt>\s*<dd>\s*(.+?)\s*</dd>',r.text,re.DOTALL).group(1),
				'visibility': re.search('<dt>visibility</dt>\s*<dd>\s*(.+?)\s*-',r.text,re.DOTALL).group(1),
				'description': re.search('<div class="blueprint-price">.*?<p>\s*(.+?)\s*</p>',r.text,re.DOTALL).group(1),
				'id': id,
			}

		# Blueprint definition
		r = bpformation.web.CallScrape("GET","/Blueprints/Designer/BlueprintXml/%s" % id)
		if r.status_code<200 or r.status_code>=300:
			bpformation.output.Status('ERROR',3,"Error retrieving data (http response %s)" % r.status_code)
			raise(bpformation.BPFormationFatalExeption("Fatal Error"))

		t = etree.XML(r.text)
		for o in t.findall("Tasks/*"):  
			try:
				bp['tasks'] += Blueprint._ExportProcessRoot(o)
			except:
					# Catch only our exceptions
				pass

		# Populate Blueprint execute stub with script-specific vals
		global_bp_values = None
		for o in t.findall("UI/Group"):  
			if o.get("Name") == "Global Blueprint Values":  
				global_bp_values =  o
				break
		if global_bp_values is not None:
			for o in global_bp_values.findall("Parameter"):  
				if o.get("Default"):  default = o.get("Default")
				elif o.get("Type") in ("Option","MultiSelect"):
					default = " | ".join([ opt.get("Value") for opt in o.findall("Option") ])
				else:  default = ''
				bp['execute'][o.get("Variable")] = default

		# Output
		if file=="-":  
			print json.dumps(bp,sort_keys=True,indent=4,separators=(',', ': '))
		else:
			if file is None:  file = "%s-%s-%s.json" % (re.sub("[^a-zA-Z0-9\-_]","_",bp['metadata']['name']).lower(),id,bp['metadata']['version'])
			bpformation.output.Status('SUCCESS',3,"%s v%s exported to %s (%s tasks)" % (bp['metadata']['name'],bp['metadata']['version'],file,len(bp['tasks'])))
			with open(file,'w') as fh:
				fh.write(json.dumps(bp,sort_keys=True,indent=4,separators=(',', ': ')))


	# Available, but not returning:
	#  o Price information
	#  o Environment size details (server count, memory, CPU, etc.)
	#  o Blueprint description
	@staticmethod
	def List(filters):
		# TODO - much of this is available with a json response as part of group execute package
		r = bpformation.web.CallScrape("POST","/blueprints/browser/LoadTemplates",payload={
					'Search-PageSize': 1000,
					'Search-PageNumber': 1,
				}).text

		blueprints = []
		for blueprint_html in filter(lambda x: x in Blueprint.limited_printable_chars, r).split('class="blueprint-specs"'):
			try:
				blueprints.append({'name': re.search('class="blueprint-header">.*<label>(.+?)</label>',blueprint_html).group(1),
								   'date_added': re.search('<em>(.+?)</em>',blueprint_html).group(1),
								   'description': re.search('<div class="blueprint-desc">\s*<div><strong>(.+?)</strong>',blueprint_html).group(1),
								   'visibility': re.search('<div class="right-col"><strong>(.+?)</strong></div>',blueprint_html).group(1),
						           'id': re.search('<a href="/blueprints/browser/details/(\d+)">',blueprint_html).group(1) })
				
			except:
				pass

		# Apply filters if any are specified
		if not filters:  blueprints_final = blueprints
		else:
			blueprints_final = []
			for blueprint in blueprints:
				match = True
				for one_filter in filters:
					if not re.search(one_filter," ".join(blueprint.values()),re.IGNORECASE):  match = False
				if  match:  blueprints_final.append(blueprint)

		return(blueprints_final)


	@staticmethod
	def _ImportAnonymizeTasks(o):
		if o['type'] == 'server' and 'uuid' in o: del(o['uuid'])
		if 'id' in o:  del(o['id'])
		if 'tasks' in o:
			for idx,task in enumerate(o['tasks']):
				o['tasks'][idx] = Blueprint._ImportAnonymizeTasks(task)

		return(o)


	@staticmethod
	def _PostServer(blueprint_id,o):
		"""
		Post to create/update server configuration.

		Must include all parameters for the server - anything not specified is "deleted".

		o is a server object with keys name, description, cpu, ram, template, id, and a list of tasks.
		o['id'] is 0 for new servers, otherwise include the uuid of the existing server to modify.
		"""

		# Validate core params
		for key in ('name','description','cpu','ram'):
			if key not in o:
				bpformation.output.Status('ERROR',3,"Blueprint json server definition missing '%s'" % key)
				raise(bpformation.BPFormationFatalExeption("Fatal Error"))
		if 'id' not in o:  o['id'] = 0	# New servers don't start with an ID
		if len(o['name'])==0 or len(o['name'])>5:
			bpformation.output.Status('ERROR',3,"Blueprint json server name must be between 0 and five characters" % key)
			raise(bpformation.BPFormationFatalExeption("Fatal Error"))
		if not re.match("^\d+$",o['cpu']) or int(o['cpu'])==0 or int(o['cpu'])>16:
			bpformation.output.Status('ERROR',3,"Blueprint json server cpu must be between 1 and 16" % key)
			raise(bpformation.BPFormationFatalExeption("Fatal Error"))
		if not re.match("^\d+$",o['ram']) or int(o['ram'])==0 or int(o['cpu'])>128:
			bpformation.output.Status('ERROR',3,"Blueprint json server ram must be between 1 and 128" % key)
			raise(bpformation.BPFormationFatalExeption("Fatal Error"))

		# Build tasks data structure
		# TODO - query for parameters and verify we meet requirements
		staged_tasks = {}
		staged_tasks_idx = 0 
		if 'tasks' in o:
			for task in o['tasks']:
				# TODO system tasks - scheduled task, delete snapshot, revert snapshot
				# System - add disk
				if task['type']=='disk' and task['uuid']=='22460210-b682-4138-93fd-1a95c5e4e039':
					staged_tasks['Server.Tasks[%s].ID' % staged_tasks_idx] = task['uuid']
					staged_tasks['Server.Tasks[%s].Properties[0].Name' % staged_tasks_idx] = 'GB'
					staged_tasks['Server.Tasks[%s].Properties[0].Value' % staged_tasks_idx] = task['gb']
					if 'drive' in task:
						staged_tasks['Server.Tasks[%s].Properties[1].Name' % staged_tasks_idx] = 'Drive'
						staged_tasks['Server.Tasks[%s].Properties[1].Value' % staged_tasks_idx] = task['drive']
	
				# System - reboot
				elif task['type']=='reboot' and task['uuid']=='5b949945-6981-4a81-bbcc-4ddd3d394b8d':
					staged_tasks['Server.Tasks[%s].ID' % staged_tasks_idx] = task['uuid']
	
				# System - add public NAT
				elif task['type']=='add_nat_ip' and task['uuid']=='c000d327-3543-4d9e-ac43-e8fbce4620ab':
					staged_tasks['Server.Tasks[%s].ID' % staged_tasks_idx] = task['uuid']
					staged_tasks['Server.Tasks[%s].Properties[0].Name' % staged_tasks_idx] = 'FirewallOptions'
					staged_tasks['Server.Tasks[%s].Properties[0].Value' % staged_tasks_idx] = ",".join(task['ingress_ports'])
	
				# System - add add'l private IP
				elif task['type']=='add_ip' and task['uuid']=='9a851f50-c676-4c11-b4c8-a0a7241c1060':
					staged_tasks['Server.Tasks[%s].ID' % staged_tasks_idx] = task['uuid']

				# System - set ttl
				#elif task['type']=='ttl' and task['uuid']=='2d4d37e6-be83-461a-bfd2-b3c2ca327db5':
				#	staged_tasks['Server.Tasks[%s].ID' % staged_tasks_idx] = task['uuid']
				#	staged_tasks['Server.Tasks[%s].Properties[0].Name' % staged_tasks_idx] = 'ActivityType'
				#	staged_tasks['Server.Tasks[%s].Properties[0].Value' % staged_tasks_idx] = 'Delete'
				#	staged_tasks['Server.Tasks[%s].Properties[1].Name' % staged_tasks_idx] = 'TimeZone'
				#	staged_tasks['Server.Tasks[%s].Properties[1].Value' % staged_tasks_idx] = 'UTC'
				#	staged_tasks['Server.Tasks[%s].Properties[2].Name' % staged_tasks_idx] = 'ScheduleDate'
				#	staged_tasks['Server.Tasks[%s].Properties[2].Value' % staged_tasks_idx] = '2015-06-06T15:35:00Z'
	
				# User package
				elif task['type']=='package':
					staged_tasks['Server.Tasks[%s].ID' % staged_tasks_idx] = task['uuid']
	
					# Add all design-time parameters
					try:
						properties_idx = 0
						for key,value in task['properties'].items():
							staged_tasks['Server.Tasks[%s].Properties[%s].Name' % (staged_tasks_idx,properties_idx)] = key
							staged_tasks['Server.Tasks[%s].Properties[%s].Value' % (staged_tasks_idx,properties_idx)] = value
							properties_idx += 1
					except:
						pass
	
				# Unknown type/ID
				else:
					bpformation.output.Status('ERROR',3,"Blueprint json server task unknown type/id '%s'" % task['type'])
					raise(bpformation.BPFormationFatalExeption("Fatal Error"))
	
				staged_tasks_idx += 1

		# Post
		r = bpformation.web.CallScrape("POST","/blueprints/designer/SaveServer?id=%s" % blueprint_id,allow_redirects=False,payload=dict({
					"Server.ID": o['id'],
					"Server.Template": o['template'],
					"Server.Name": o['name'],
					"Server.Description": o['description'],
					"Server.Processor": o['cpu'],
					"Server.Memory": o['ram'],
				}.items() + staged_tasks.items()))
		if r.status_code<200 or r.status_code>=300:
			bpformation.output.Status('ERROR',3,"Error creating blueprint - step 2 add server failure (response %s)" % r.status_code)
			raise(bpformation.BPFormationFatalExeption("Fatal Error"))

		if o['id']==0:  o['id'] = r.json()['serverID']
		return(o)


	@staticmethod
	def _PostBlueprint(bp):
		"""Update and publish provided Blueprint objects."""

		# Validate syntax and required metadata fields
		for key in ('description','name','visibility','version'):
			if key not in bp['metadata']:
				bpformation.output.Status('ERROR',3,"Blueprint json missing metadata/%s" % key)
				raise(bpformation.BPFormationFatalExeption("Fatal Error"))

		# Munge version formatting
		if not re.search("^[\d\.]+$",bp['metadata']['version']):  
			bpformation.output.Status('ERROR',3,"Blueprint json version must contain only a dotted number representation")
			raise(bpformation.BPFormationFatalExeption("Fatal Error"))
		m = re.search("(.+?)\.?(.*)",bp['metadata']['version'])
		ver_major = int(m.group(1))
		if len(m.groups(2)):  ver_minor = int(m.group(2).replace(".",""))
		else:  ver_minor = 0

		# Step 0 - Get UUID if existing
		if bp['metadata']['id']:
			r = bpformation.web.CallScrape("GET","/blueprints/browser/details/%s" % bp['metadata']['id'])
			if r.status_code<200 or r.status_code>=300:
				bpformation.output.Status('ERROR',3,"Error creating blueprint - step 0 uuid failure (response %s)" % r.status_code)
				raise(bpformation.BPFormationFatalExeption("Fatal Error"))
			bp['metadata']['uuid'] = re.search('/Blueprints/Browser/Clone.blueprintUUID=([a-zA-Z0-9\-]+)',r.text,re.DOTALL).group(1)
			r = bpformation.web.CallScrape("GET","/blueprints/designer/metadata/%s" % bp['metadata']['uuid'],allow_redirects=False)
		else:  bp['metadata']['uuid'] = str(uuid.uuid4())

		# Step 1 - Metadata post and create Blueprint shell
		r = bpformation.web.CallScrape("POST","/blueprints/designer/metadata",allow_redirects=False,payload={
					"capabilities": "",     # Aligns to "tags"
					"companySize": 3,       # 1: 1-100, 2: 101-1,000, 3: 1001-5000, 4: 5,000+
					"isReseller": False,    # Tied to is_managed, no-op
					"errors": [],           # unknown
					"userCapabilities": "", # TODO - Custom tags
					"templateID": bp['metadata']['id'],
					"description": bp['metadata']['description'],
					"templateName": bp['metadata']['name'],
					"visibility": Blueprint.visibility_stoi[bp['metadata']['visibility'].lower()],
					"versionMajor": ver_major,
					"versionMinor": ver_minor,
				})
		if r.status_code<200 or r.status_code>=300:
			bpformation.output.Status('ERROR',3,"Error creating blueprint - step 1 metadata failure (response %s)" % r.status_code)
			raise(bpformation.BPFormationFatalExeption("Fatal Error"))
		bp['metadata']['id'] = re.search("(\d+)$",r.json()['url']).group(1)

		# Step 2 - Apply all tasks
		new_tasks = []
		for task in bp['tasks']:
			if task['type'] == 'server':  new_tasks.append(Blueprint._PostServer(bp['metadata']['id'],task))
			else:
				bpformation.output.Status('ERROR',3,"Unknown task type '%s'" % task['type'])
				raise(bpformation.BPFormationFatalExeption("Fatal Error"))
		bp['tasks'] = new_tasks

		# Step 3 - Publish Blueprint
		r = bpformation.web.CallScrape("POST","/Blueprints/Designer/review/%s" % bp['metadata']['id'],allow_redirects=False,payload={
					"Publish": "", # unknown
					"DataTemplate.UUID": bp['metadata']['uuid'],
					"TemplateID": bp['metadata']['id'],
				})

		# TODO Step 4 - save output?  Assume needed for some kind of update

		if bp['metadata']['visibility'].lower()=="public":  
				bpformation.output.Status('SUCCESS',3,"CenturyLink approval needed for Public publishing.  Email ecosystem@CenturyLinkCloud.com for approval")

		return(bp)


	@staticmethod
	def Import(files):
		for file in files:
			if not os.path.isfile(file):
				bpformation.output.Status('ERROR',3,"Blueprint json file '%s' not found" % file)

			# Load json
			with open(file) as fh:
				bp = json.load(fh)

			# Strip unique IDs from assets that will be duplicated
			bp['metadata']['id'] = 0
			new_tasks = []
			for task in bp['tasks']:  new_tasks.append(Blueprint._ImportAnonymizeTasks(task))
			bp['tasks'] = new_tasks
				
			# Publish BP, obtain BP with ids as result
			bp = Blueprint._PostBlueprint(bp)
			bpformation.output.Status('SUCCESS',3,"%s v%s imported ID %s (%s tasks)" % (bp['metadata']['name'],bp['metadata']['version'],bp['metadata']['id'],len(bp['tasks'])))
			#bpformation.output.Status('SUCCESS',3,"Blueprint created with ID %s (https://control.ctl.io/blueprints/browser/details/%s)" % (blueprint_id,blueprint_id))


	@staticmethod
	def Update(files):
		for file in files:
			if not os.path.isfile(file):
				bpformation.output.Status('ERROR',3,"Blueprint json file '%s' not found" % file)

			# Load json
			with open(file) as fh:
				bp = json.load(fh)

			# Publish BP, obtain BP with ids as result
			bp = Blueprint._PostBlueprint(bp)
			bpformation.output.Status('SUCCESS',3,"%s v%s updated ID %s (%s tasks)" % (bp['metadata']['name'],bp['metadata']['version'],bp['metadata']['id'],len(bp['tasks'])))
			#bpformation.output.Status('SUCCESS',3,"Blueprint created with ID %s (https://control.ctl.io/blueprints/browser/details/%s)" % (blueprint_id,blueprint_id))


	@staticmethod
	def Execute(ids,files,parameters,type,password,group_id,network,dns):
		"""

		parameters - dict of key/value pairs or a list of key=value pairs
		"""

		# Build list of file-based assets to exec
		bps = []
		if files is not None and len(files):
			for file in files:
				if not os.path.isfile(file):
					bpformation.output.Status('ERROR',3,"Blueprint json file '%s' not found" % file)
	
				# Load json
				with open(file) as fh:
					bp = json.load(fh)
	
				if 'id' not in bp['metadata']:
					bpformation.output.Status('ERROR',3,"No blueprint ID in '%s'" % file)
					raise(bpformation.BPFormationFatalExeption("Fatal Error"))
	
				bps.append({'execute': bp['execute'], 'id': bp['metadata']['id']})


		# Build list of id-based assets to exec
		if ids is not None and len(ids):
			for id in ids:  bps.append({'execute': {}, 'id': id})

		# Build execute parameter lists
		# order of precedence (low to high):
		#   o .bpformation ini file
		#   o blueprint json file 'execute' obj
		#   o command line args
		new_bps = []
		for bp in bps:
			# Apply defaults from config file - don't overwrite existing items.  Only covers system items
			for key in ('type','password','group_id','network','dns'):
				if key in bp['execute']:  continue
				if bpformation.config and bpformation.config.has_option('blueprint_execute',key):  bp['execute'][key] = bpformation.config.get('blueprint_execute',key)

			# Apply system-level command line args
			for key in ('type','password','group_id','network','dns'):
				if key in vars() and vars()[key] is not None:  bp['execute'][key] = vars()[key]

			# Apply parameter command line args
			if parameters is not None and isinstance(parameters,dict):
				for key,value in parameters.items():  bp['execute'][key] = value
			elif parameters is not None and len(parameters):
				for parameter in parameters:
					(key,value) = parameter.split("=",1)
					bp['execute'][key] = value

			# Hardcoded default DNS - this is dumb that we ask for it
			if 'dns' not in bp['execute'] or not len(bp['execute']['dns']):
				bp['execute']['dns'] = "172.17.1.26"

			# Confirm system parameters are all set
			for key in ('type','password','group_id','network','dns'):
				if key not in bp['execute'] or not len(bp['execute'][key]):
					bpformation.output.Status('ERROR',3,"Missing required system parameter '%s'" % key)
					raise(bpformation.BPFormationFatalExeption("Fatal Error"))

			# TODO Confirm all blueprint parameters are set

			# Rename system parameters so they match what POST is expecting
			bp['execute']['TemplateID'] = bp['id']
			bp['execute']['T3.BuildServerTask.Password'] = bp['execute']['password']
			bp['execute']['Confirm.T3.BuildServerTask.Password'] = bp['execute']['password']
			bp['execute']['T3.BuildServerTask.GroupID'] = bp['execute']['group_id']
			bp['execute']['T3.BuildServerTask.Network'] = bp['execute']['network']
			bp['execute']['T3.BuildServerTask.PrimaryDNS'] = 'Manual'
			bp['execute']['T3.BuildServerTask.PrimaryDNS_manual'] = bp['execute']['dns']
			bp['execute']['T3.BuildServerTask.SecondaryDNS'] = ''
			bp['execute']['T3.BuildServerTask.SecondaryDNS_manual'] = ''
			bp['execute']['T3.BuildServerTask.HardwareType'] = bp['execute']['type']
			bp['execute']['T3.BuildServerTask.AntiAffinityPoolId'] = ''
			bp['execute']['T3.BuildServerTask.ServiceLevel'] = 'Standard'
			bp['execute']['RequestID'] = ''
			bp['execute']['Submit'] = ''
			for key in ('password','group_id','network','dns','type'):  del(bp['execute'][key])

			new_bps.append(bp)
				
		bps = new_bps

		# Execute each Blueprint
		clc.v2.SetCredentials(bpformation.CONTROL_USER, bpformation.CONTROL_PASSWORD)
		requests = []
		start_time = time.time()
		for bp in bps:
	
			# Step 1 - customize blueprint
			r = bpformation.web.CallScrape("POST","/Blueprints/Builder/Customize/%s" % bp['id'],allow_redirects=False,payload=bp['execute'])
			if r.status_code<200 or r.status_code>=400:
				bpformation.output.Status('ERROR',3,"Error executing blueprint - step 1 customization metadata failure (response %s)" % r.status_code)
				raise(bpformation.BPFormationFatalExeption("Fatal Error"))


			# Step 2 - deploy blueprint
			r = bpformation.web.CallScrape("POST",r.headers['location'],allow_redirects=False,payload={'TemplateID': bp['id'], 'Submit': ''})
			if r.status_code<200 or r.status_code>=400:
				bpformation.output.Status('ERROR',3,"Error executing blueprint - step 2 submit failure (response %s)" % r.status_code)
				raise(bpformation.BPFormationFatalExeption("Fatal Error"))

			# Step 3 - queue request id
			request_id = re.sub(".*/(\d+).*=(.*)",r'\2-\1',r.headers['location']).lower()
			requests.append(clc.v2.Requests([{'isQueued': True, 'links': [{'rel': 'status', 'id': request_id}]}],alias=bpformation.web.Alias()))
			bpformation.output.Status('SUCCESS',3,"Execution request submitted for Blueprint ID %s (request %s)" % (bp['id'],request_id))

		# Wait for executing blueprints to complete
		# TODO - async option
		requests = sum(requests)
		requests.WaitUntilComplete()
		if len(requests.success_requests):
			# TODO - parse request page and pull out list of servers 
			#bpformation.output.Status('SUCCESS',3,"Execution completed on %s (%s seconds)" % (",".join(success_servers),int(time.time()-start_time)))
			bpformation.output.Status('SUCCESS',3,"Execution completed on %s blueprints (%s seconds)" % (len(requests.success_requests),int(time.time()-start_time)))
		for request in requests.error_requests:
			(req_loc,req_id) = request.id.split("-",1)
			r = bpformation.web.CallScrape("GET","/Blueprints/Queue/RequestDetails/%s?location=%s" % (req_id,req_loc))
			if r.status_code<300 and r.status_code>=200:
				error = re.search('<div class="module-body">.*?<pre>(\s*.Error.\s*)?(.*?)\s*</pre>',r.text,re.DOTALL).group(2)
				bpformation.output.Status('ERROR',3,"Execution failed on %s: %s" % (request.id,error))
			else:
				bpformation.output.Status('ERROR',3,"Execution failed on %s request ID %s (https://control.ctl.io/Blueprints/Queue/RequestDetails/%s?location=%s)" % \
						(request.data['context_val'],req_id,req_id,req_loc))

		if requests.error_requests:  raise(bpformation.BPFormationFatalExeption("Error executing blueprint"))

		# TODO - percolate error up so we exit with an error called as CLI or see exception as sdk





# -*- coding: utf-8 -*-
"""
Manage Blueprint packages.
"""

import os
import re
import clc
import time
import ftplib
import requests
import bpformation


#
# TODO vCur:
#  o Query for all parameters associated with a package.  Optionally enforce on package Execute
#


#
# TODO vNext:
#  o Create new ftp endpoint if none exists
#  o Retrive list of existing packages using v2 json endpoint
#


class Package():

	visibility_stoi = { 'Public': 1, 'Private': 2, 'Shared': 3}
	limited_printable_chars = "0123456789abcdefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXYZ!\"#$%&'()*+,-./:;<=>?@[]^_`{|}~  "


	@staticmethod
	def _GetFtpEndpoint():
		# TODO - alert on failure to locate.  Maybe add scrape POST call to create if missing?

		# Scrape from ftp users page
		#text = bpformation.web.CallScrape("GET","/Blueprints/ftpuser").text
		#m = re.search("%s_BlueprintUser</td>.*?<td>(.*?)</td>.*?<td>(.*?)</td>" % bpformation.web.Alias(),text,re.DOTALL)

		#if m:
		#	bpformation.FTP_ENDPOINT = {'user': "%s_BlueprintUser" % bpformation.web.Alias(), 
		#	                            'passwd': m.group(1), 'endpoint': m.group(2)}
		#	bpformation.output.Status('SUCCESS',1,"FTP endpoint %s@%s cached" % (bpformation.FTP_ENDPOINT['user'],bpformation.FTP_ENDPOINT['endpoint']))
		#else:  bpformation.output.Status('ERROR',3,"Unable to retrieve FTP endpoint.  Make sure %s_BlueprintUser exists" % bpformation.web.Alias())
		# json response direct form scripts page

		r = bpformation.web.CallScrape("POST","/blueprints/packages/GetFTPDetails").json()
		bpformation.FTP_ENDPOINT = {'user': r['userName'], 'passwd': r['password'], 'endpoint': r['ftpHost']}
		bpformation.output.Status('SUCCESS',1,"FTP endpoint %s@%s cached" % (bpformation.FTP_ENDPOINT['user'],bpformation.FTP_ENDPOINT['endpoint']))



	@staticmethod
	def Upload(files):
		for file in files:
			if not os.path.isfile(file):
				bpformation.output.Status('ERROR',3,"Package file '%s' not found" % file)
				raise(bpformation.BPFormationFatalExeption("Fatal Error"))

		if not bpformation.FTP_ENDPOINT:  Package._GetFtpEndpoint()

		# TODO - alert nicely on login failure
		# TODO - output xfer stats if verbose
		ftp = ftplib.FTP(bpformation.FTP_ENDPOINT['endpoint'],bpformation.FTP_ENDPOINT['user'],bpformation.FTP_ENDPOINT['passwd'])
		for file in files:
			time_start = time.time()
			with open(file,'rb') as fh:
				file_name = re.sub(".*/","",file)
				ftp.storbinary("STOR %s" % (file_name),fh)
			bpformation.output.Status('SUCCESS',3,"%s successfully uploaded (%s seconds)" % (file_name,int(time.time()-time_start)))
		ftp.quit()

	
	@staticmethod
	def _PackageOSAtoI(type,os_regexs):
		# TODO - catch error
		oss = bpformation.web.CallScrape("POST","/blueprints/packages/GetOSList", {'osType': type}).json()['Result']

		oss_retain = {}
		for os_regex in os_regexs:  
			if re.match("rhel",os_regex.lower()):  os_regex = "RedHat"
			for os in oss:  
				if re.search(os_regex,os['Name']):  oss_retain[os['ID']] = True

		return(oss_retain.keys())

	
	@staticmethod
	def ListOS(type):
		try:
			return(bpformation.web.CallScrape("POST","/blueprints/packages/GetOSList", {'osType': type}).json()['Result'])
		except:
			bpformation.output.Status('ERROR',3,"Unable to retrieve OS listi")
			raise(bpformation.BPFormationFatalExeption("Fatal Error"))


	@staticmethod
	def Publish(files,type,visibility,os):
		task_queue = []
		for file in files:
			r = bpformation.web.CallScrape("POST","/Blueprints/Packages/Properties",
							               payload={"packageName": file,
									                "Type": 2,
									                "OS": type,
									                "Permissions": Package.visibility_stoi[visibility],
									                "OS_Version": Package._PackageOSAtoI(type,os)})
			if r.status_code>=200 and r.status_code<400:
				m = re.search("<a href=\"/Blueprints/Queue/RequestDetails/(\d+)\?location=(.+)\"",r.text)
				task_queue.append({'id': int(m.group(1)), 'location': m.group(2), 'description': file, 'date_added': time.time()})
				bpformation.output.Status('SUCCESS',3,"%s publish job submitted" % file)
			elif re.search("Unable to publish software package. .*A new UUID is required",r.text):
				bpformation.output.Status('ERROR',3,"Unable to publish %s. A new UUID is required" % file)
			else:
				bpformation.output.Status('ERROR',3,"Unknown error publishing %s (http response code %s)" % (file,r.status_code))
				

		bpformation.queue.WaitForQueue(task_queue)

		if visibility.lower()=="public":
			bpformation.output.Status('SUCCESS',3,"CenturyLink approval needed for Public publishing.  Email ecosystem@CenturyLinkCloud.com for approval")

	
	@staticmethod
	def Delete(uuids):
		for uuid in uuids:
			r = bpformation.web.CallScrape("POST","/blueprints/packages/DeletePackage/",
							               payload={"id": uuid, "classification": "Script", })
			if r.status_code<400 and r.status_code>=200:
				bpformation.output.Status('SUCCESS',3,"%s package deleted" % uuid)
			else:
				bpformation.output.Status('ERROR',3,"%s package deletion error (status code %s)" % (uuid,r.status_code))

	
	@staticmethod
	def Download(uuids):
		if not bpformation._CONTROL_COOKIES:  bpformation.web._LoginScrape()

		for uuid in uuids:
			r = requests.get("%s/Blueprints/Packages/Download?uuid=%s" % (bpformation.defaults.CONTROL_URL,uuid), 
			                 cookies=bpformation._CONTROL_COOKIES,
							 stream=True)
			with open("%s.zip" % uuid, 'wb') as f:
				for chunk in r.iter_content(chunk_size=1024):
					if chunk: # filter out keep-alive new chunks
						f.write(chunk)
						f.flush()
			bpformation.output.Status('SUCCESS',3,"%s package downloaded" % uuid)

	
	@staticmethod
	def List(filters):
		# TODO - Query json v2 endpoint - https://control.ctl.io/api/tunnel/v2/packages/krap
		r = bpformation.web.CallScrape("GET","/Blueprints/packages/Library").text
		table = re.search('id="PackageLibrary">.*?<table class="table">.*?<tbody>(.*)</tbody>',r,re.DOTALL).group(1)

		packages = []
		for package_html in filter(lambda x: x in Package.limited_printable_chars, table).split("</tr>"):
			#m = re.search('<td>\s*(.+?)\s*<input id="package_UUID".*?value="(.+?)".*?<td>(.*?)</td>.*?<td>.*(.+?)</td>i.*</i>(.*?)</td>.*<td>.*\s(.+?)\s*</td>',package_html,re.DOTALL)
			cols = package_html.split("<td>")
			try:
				packages.append({'name': re.search('\s*(.+?)\s*<input',cols[1],re.DOTALL).group(1),
				                 'uuid': re.search('<input id="package_UUID" name="package.UUID" type="hidden" value="(.*?)"\s*/>',cols[1],re.DOTALL).group(1),
						         'owner': re.search('(.*?)</td>',cols[2]).group(1),
						         'status': re.search('</i>(.+?)</td>',cols[3]).group(1),
						         'visibility': re.search('\s*(.+?)\s*</td>',cols[4]).group(1) })
				
			except:
				pass
			#print m.group(0)
			#if m:  packages.append({'name': m.group(1), 'uuid': m.group(2), 'owner': m.group(3), 'visibility': m.group(4)})

		# Apply filters if any are specified
		if not filters:  packages_final = packages
		else:
			packages_final = []
			for package in packages:
				match = True
				for one_filter in filters:
					if not re.search(one_filter," ".join(package.values()),re.IGNORECASE):  match = False
				if  match:  packages_final.append(package)

		return(packages_final)


	@staticmethod
	def Execute(uuid,servers,parameters):
		# TODO - async option
		clc.v2.SetCredentials(bpformation.CONTROL_USER, bpformation.CONTROL_PASSWORD)
		requests = []
		start_time = time.time()
		for server in servers:
			bpformation.output.Status('SUCCESS',3,"Execution request submitted for %s" % server.lower())
			if parameters:  parameters = dict( (p.split("=")[0],p.split("=",1)[1]) for p in parameters )
			else:  parameters=None
			requests.append(clc.v2.Server(server,alias=bpformation.web.Alias()).ExecutePackage(
					package_id=re.sub("[^a-zA-Z0-9]","",uuid).lower(),   # Wants uuid w/o dashes
					parameters=parameters,
				))

		requests = sum(requests)
		requests.WaitUntilComplete()
		if len(requests.success_requests):
			success_servers = [ o.data['context_val'] for o in requests.success_requests ]
			bpformation.output.Status('SUCCESS',3,"Execution completed on %s (%s seconds)" % (", ".join(success_servers),int(time.time()-start_time)))
		for request in requests.error_requests:
			(req_loc,req_id) = request.id.split("-",1)
			r = bpformation.web.CallScrape("GET","/Blueprints/Queue/RequestDetails/%s?location=%s" % (req_id,req_loc))
			if r.status_code<300 and r.status_code>=200:
				error = re.search('<div class="module-body">.*?<pre>(\s*.Error.\s*)?(.*?)\s*</pre>',r.text,re.DOTALL).group(2)
				bpformation.output.Status('ERROR',3,"Execution failed on %s: %s" % (request.data['context_val'],error))
			else:
				bpformation.output.Status('ERROR',3,"Execution failed on %s request ID %s (https://control.ctl.io/Blueprints/Queue/RequestDetails/%s?location=%s)" % \
						(request.data['context_val'],req_id,req_id,req_loc))






# -*- coding: utf-8 -*-
"""
CenturyLink Cloud Python Blueprint Package and Workflow CLI management tool.

CenturyLink Cloud: http://www.CenturyLinkCloud.com
Package Github page: https://github.com/CenturyLinkCloud/bp-formation

"""

from bpformation.shell import Args, ExecCommand
from bpformation.package import Package as package
from bpformation.blueprint import Blueprint as blueprint
from bpformation.web import Web as web
from bpformation.api import API as api
from bpformation.queue import Queue as queue
import bpformation.output
import defaults


####### module/object vars #######
CONTROL_USER = False
CONTROL_PASSWORD = False
ALIAS = False
FTP_ENDPOINT = False


__version__ = "0.8"

args = False
config = False

_SSL_VERIFY = True

_ALIAS = False
_LOGINS = 0
_CONTROL_COOKIES = False
_BLUEPRINT_FTP_URL = False

class BPFormationExeption(Exception):
	pass
class BPFormationFatalExeption(Exception):
	pass
class BPFormationLoginException(BPFormationExeption):
	pass




#!/usr/bin/env python
import os
import shlex
import struct
import platform
import subprocess
 
 
def get_terminal_size():
	""" getTerminalSize()
	 - get width and height of console
	 - works on linux,os x,windows,cygwin(windows)
	 originally retrieved from:
	 http://stackoverflow.com/questions/566746/how-to-get-console-window-width-in-python
	"""
	current_os = platform.system()
	tuple_xy = None
	if current_os == 'Windows':
		tuple_xy = _get_terminal_size_windows()
		if tuple_xy is None:
			tuple_xy = _get_terminal_size_tput()
			# needed for window's python in cygwin's xterm!
	if current_os in ['Linux', 'Darwin'] or current_os.startswith('CYGWIN'):
		tuple_xy = _get_terminal_size_linux()
	if tuple_xy is None:
		print "default"
		tuple_xy = (80, 25)	  # default value
	return tuple_xy
 
 
def _get_terminal_size_windows():
	try:
		from ctypes import windll, create_string_buffer
		# stdin handle is -10
		# stdout handle is -11
		# stderr handle is -12
		h = windll.kernel32.GetStdHandle(-12)
		csbi = create_string_buffer(22)
		res = windll.kernel32.GetConsoleScreenBufferInfo(h, csbi)
		if res:
			(bufx, bufy, curx, cury, wattr,
			 left, top, right, bottom,
			 maxx, maxy) = struct.unpack("hhhhHhhhhhh", csbi.raw)
			sizex = right - left + 1
			sizey = bottom - top + 1
			return sizex, sizey
	except:
		pass
 
 
def _get_terminal_size_tput():
	# get terminal width
	# src: http://stackoverflow.com/questions/263890/how-do-i-find-the-width-height-of-a-terminal-window
	try:
		cols = int(subprocess.check_call(shlex.split('tput cols')))
		rows = int(subprocess.check_call(shlex.split('tput lines')))
		return (cols, rows)
	except:
		pass
 
 
def _get_terminal_size_linux():
	def ioctl_GWINSZ(fd):
		try:
			import fcntl
			import termios
			cr = struct.unpack('hh',
							   fcntl.ioctl(fd, termios.TIOCGWINSZ, '1234'))
			return cr
		except:
			pass
	cr = ioctl_GWINSZ(0) or ioctl_GWINSZ(1) or ioctl_GWINSZ(2)
	if not cr:
		try:
			fd = os.open(os.ctermid(), os.O_RDONLY)
			cr = ioctl_GWINSZ(fd)
			os.close(fd)
		except:
			pass
	if not cr:
		try:
			cr = (os.environ['LINES'], os.environ['COLUMNS'])
		except:
			return None
	return int(cr[1]), int(cr[0])
 
if __name__ == "__main__":
	sizex, sizey = get_terminal_size()
	print  'width =', sizex, 'height =', sizey



"""Command-line interface to the CenturyLink Cloud (CLC) API."""

import re
import os
import sys
import argparse
import ConfigParser

import bpformation


class Args:

	def __init__(self):
		bpformation.args = self
		self.ParseArgs()
		self.ImportIni()
		self.MergeEnvironment()
		self.MergeCommands()


	def ParseArgs(self):
		parser = argparse.ArgumentParser(description="CLI tool for interacting with CenturyLink Blueprints Package and Workflow Services.  http://www.CenturyLinkCloud.com (v%s)" % bpformation.__version__)
		parser_sp1 = parser.add_subparsers(title='Commands',dest='command')

		########## Package ###########
		#
		# TODO vCur:
		#
		# TODO vNext:
		#  o dependencies tree
		#  o package (zip given file manifest or directory.  Gen UUID.  Multiple options based on publish target)
		#

		parser_package = parser_sp1.add_parser('package', help='Package level activities (list, package, upload, publish, etc.)')
		parser_sp2 = parser_package.add_subparsers(dest='sub_command')

		## List
		parser_package_list = parser_sp2.add_parser('list', help='List package inventory')
		parser_package_list.add_argument('--filter', nargs='*', required=False, help='Regex filter Results by name, author, status, visibility (and)')

		## Upload
		parser_package_upload = parser_sp2.add_parser('upload', help='Uploaded package to specified alias')
		parser_package_upload.add_argument('--file', nargs='*', metavar="FILE", required=True, help='Files to upload')

		## List OS
		parser_package_listos = parser_sp2.add_parser('list-os', help='List known operating systems')
		parser_package_listos.add_argument('--type', required=True, choices=['Windows','Linux'], help='Operating system')

		## Publish
		parser_package_publish = parser_sp2.add_parser('publish', help='Uploaded packages to specified alias')
		parser_package_publish.add_argument('--file', nargs='*', required=True, help='Uploaded filename to process')
		parser_package_publish.add_argument('--type', required=True, choices=['Windows','Linux'], help='Operating system')
		parser_package_publish.add_argument('--visibility', required=True, choices=['Public','Private','Shared'], help='Package visibility')
		parser_package_publish.add_argument('--os', nargs='*', required=True, help='Operating system list (regex supported)')

		# Upload and publish
		parser_package_uploadpublish= parser_sp2.add_parser('upload-and-publish', help='Uploaded then publish packages to specified alias')
		parser_package_uploadpublish.add_argument('--file', nargs='*', metavar="FILE", required=True, help='Files to process')
		parser_package_uploadpublish.add_argument('--type', required=True, choices=['Windows','Linux'], help='Operating system')
		parser_package_uploadpublish.add_argument('--visibility', required=True, choices=['Public','Private','Shared'], help='Package visibility')
		parser_package_uploadpublish.add_argument('--os', nargs='*', required=True, help='Operating system list (regex supported)')

		## Delete
		parser_package_delete = parser_sp2.add_parser('delete', help='Delete published package')
		parser_package_delete.add_argument('--uuid', nargs='*', required=True, help='UUID for packages')

		## Download
		parser_package_download = parser_sp2.add_parser('download', help='Download published package')
		parser_package_download.add_argument('--uuid', nargs='*', required=True, help='UUID for packages')

		## Execute
		parser_package_execute = parser_sp2.add_parser('execute', help='Execute package')
		parser_package_execute.add_argument('--uuid', required=True, help='UUID for package')
		parser_package_execute.add_argument('--server', required=True, nargs='*', help='Servers targets for package execution')
		parser_package_execute.add_argument('--parameter', nargs='*', help='key=value pairs for package parameters')



		########## Blueprint ###########
		#
		# TODO vCur:
		#
		# TODO vNext:
		#  o create (interactive ui to build blueprint)
		#  o dependencies tree
		#

		parser_blueprint = parser_sp1.add_parser('blueprint', help='Blueprint level activities (list, import, export, delete, etc.)')
		parser_sp2 = parser_blueprint.add_subparsers(dest='sub_command')

		## List
		parser_blueprint_list = parser_sp2.add_parser('list', help='List blueprint inventory')
		parser_blueprint_list.add_argument('--filter', nargs='*', required=False, help='Regex filter Results by name, author, status, visibility (and)')

		## Export
		parser_blueprint_export = parser_sp2.add_parser('export', help='Export blueprint')
		parser_blueprint_export.add_argument('--id', required=True, help='Blueprint ID (note this ID is not globally unique - find this from your primary datacenter')
		parser_blueprint_export.add_argument('--file', required=False, help='Filename target for Blueprint json definition')

		## Import
		parser_blueprint_import = parser_sp2.add_parser('import', help='Import blueprint from json')
		parser_blueprint_import.add_argument('--file', nargs='*', required=True, help='Blueprint definition json files')

		## Update
		parser_blueprint_update = parser_sp2.add_parser('update', help='Update existing blueprint from json')
		parser_blueprint_update.add_argument('--file', nargs='*', required=True, help='Blueprint definition json files')

		## Delete
		parser_blueprint_delete = parser_sp2.add_parser('delete', help='Delete blueprint')
		parser_blueprint_delete.add_argument('--id', nargs='*', required=True, help='Blueprint ID (note this ID is not globally unique - find this from your primary datacenter')

		## Execute
		# TODO - update verbiage/grouping to make less confusing.  User can specify any combo of file, id, parameter, etc.
		#        as long as in the end they can be sensibly merged into a unified view that supplies all needed info.
		parser_blueprint_execute = parser_sp2.add_parser('execute', help='Execute blueprint')
		parser_blueprint_execute.add_argument('--file', nargs='*', help='Blueprint definition json files with "execute" populated')
		parser_blueprint_execute.add_argument('--id', nargs='*', help='Blueprint ID (note this ID is not globally unique - find this from your primary datacenter')
		parser_blueprint_execute.add_argument('--parameter', nargs='*', help='key=value pairs for package parameters (overrides "file")')
		parser_blueprint_execute.add_argument('--type', choices=['Standard','HyperScale'], help='Server hardware type')
		parser_blueprint_execute.add_argument('--password', help='Server deploy password')
		parser_blueprint_execute.add_argument('--group-id', help='Server deploy group ID')
		parser_blueprint_execute.add_argument('--network', help='Server deploy network')
		parser_blueprint_execute.add_argument('--dns', help='Server DNS')



		########## Global ###########
		#parser.add_argument('--cols', nargs='*', metavar='COL', help='Include only specific columns in the output')
		parser.add_argument('--config', '-c', help='Ini config file')
		parser.add_argument('--alias', '-a', help='Operate on specific account alias')
		parser.add_argument('--control-user', metavar='USER', help='Control username')
		parser.add_argument('--control-password', metavar='PASS', help='Control password')
		parser.add_argument('--quiet', '-q', default=1, action='count', help='Supress status output (repeat up to 2 times)')
		parser.add_argument('--verbose', '-v', action='count', help='Increase verbosity')
		parser.add_argument('--cols', nargs='*', metavar='COL', help='Include only specific columns in the output')
		parser.add_argument('--format', '-f', choices=['json','table','text','csv'], default='table', help='Output result format (table is default)')
		self.args = parser.parse_args()


	def GetCommand(self):  return(self.args.command)
	def GetArgs(self):  return(self.args)


	def ImportIni(self):
		config_file = False
		# Order of preference - cmd line specified, home directory file, or system file
		if self.args.config:
			config_file = self.args.config
			if self.args.config and not os.path.isfile(self.args.config):
				bpformation.output.Status('ERROR',3,"Config file %s not found" % (self.args.config))
				sys.exit(1)
		elif os.name=='nt':
			if os.path.isfile("%s/bpformation/bpformation.ini" % (os.getenv("PROGRAMDATA"))):
				config_file = "%s/bpformation/bpformation.ini" % (os.getenv("PROGRAMDATA"))
			elif os.path.isfile("bpformation.ini"):
				config_file = "bpformation.ini"
		else:
			if os.path.isfile("%s/.bpformation" % (os.environ['HOME'])):
				config_file = "%s/.bpformation" % (os.environ['HOME'])
			elif os.path.isfile("/usr/local/etc/bpformation_config"):
				config_file = "/usr/local/etc/bpformation_config"
		if config_file:  
			bpformation.output.Status('SUCCESS',0,"Reading %s" % (config_file))
			bpformation.config = ConfigParser.ConfigParser()
			bpformation.config.read(config_file)

			if bpformation.config.has_option('global','control_user'):  bpformation.CONTROL_USER = bpformation.config.get('global','control_user')
			if bpformation.config.has_option('global','control_password'):  bpformation.CONTROL_PASSWORD = bpformation.config.get('global','control_password')


	def MergeEnvironment(self):
		if 'CONTROL_USER' in os.environ:  bpformation.CONTROL_USER = os.environ['CONTROL_USER']
		if 'CONTROL_PASSWORD' in os.environ:  bpformation.CONTROL_PASSWORD = os.environ['CONTROL_PASSWORD']


	def MergeCommands(self):
		if self.args.control_user:  bpformation.CONTROL_USER = self.args.control_user
		if self.args.control_password:  bpformation.CONTROL_PASSWORD = self.args.control_password




class ExecCommand():
	def __init__(self):
		try:
			self.Bootstrap()
		except Exception as e:
			bpformation.output.Status("ERROR",3,"Exiting due to error: %s" % (str(e)))
			sys.exit(1)


	def Bootstrap(self):
		if bpformation.args.GetCommand() == 'package':  self.Package()
		elif bpformation.args.GetCommand() == 'blueprint':  self.Blueprint()


	def Package(self):
		if bpformation.args.GetArgs().sub_command == 'upload':  self.PackageUpload()
		elif bpformation.args.GetArgs().sub_command == 'list-os':  self.PackageListOS()
		elif bpformation.args.GetArgs().sub_command == 'publish':  self.PackagePublish()
		elif bpformation.args.GetArgs().sub_command == 'upload-and-publish':  self.PackageUploadAndPublish()
		elif bpformation.args.GetArgs().sub_command == 'delete':  self.PackageDelete()
		elif bpformation.args.GetArgs().sub_command == 'download':  self.PackageDownload()
		elif bpformation.args.GetArgs().sub_command == 'list':  self.PackageList()
		elif bpformation.args.GetArgs().sub_command == 'execute':  self.PackageExecute()


	def Blueprint(self):
		if bpformation.args.GetArgs().sub_command == 'list':  self.BlueprintList()
		elif bpformation.args.GetArgs().sub_command == 'export':  self.BlueprintExport()
		elif bpformation.args.GetArgs().sub_command == 'delete':  self.BlueprintDelete()
		elif bpformation.args.GetArgs().sub_command == 'import':  self.BlueprintImport()
		elif bpformation.args.GetArgs().sub_command == 'update':  self.BlueprintUpdate()
		elif bpformation.args.GetArgs().sub_command == 'execute':  self.BlueprintExecute()


	def PackageUpload(self):
		self.Exec('bpformation.package.Upload', {'files': bpformation.args.args.file}, cols=[])


	def PackagePublish(self):
		self.Exec('bpformation.package.Publish', {'files': bpformation.args.args.file, 'type': bpformation.args.args.type,
		                                          'visibility': bpformation.args.args.visibility, 'os': bpformation.args.args.os}, cols=[])


	def PackageUploadAndPublish(self):
		# Upload
		self.PackageUpload()

		# Publish
		bpformation.args.args.file = [re.sub(".*/","", o) for o in bpformation.args.args.file ]
		self.PackagePublish()


	def PackageDelete(self):
		self.Exec('bpformation.package.Delete', {'uuids': bpformation.args.args.uuid }, cols=[])


	def PackageDownload(self):
		self.Exec('bpformation.package.Download', {'uuids': bpformation.args.args.uuid }, cols=[])


	def PackageList(self):
		self.Exec('bpformation.package.List', {'filters': bpformation.args.args.filter }, cols=['name','uuid','owner','visibility','status'])


	def PackageListOS(self):
		self.Exec('bpformation.package.ListOS', {'type': bpformation.args.args.type }, cols=['Name','ID'])


	def PackageExecute(self):
		self.Exec('bpformation.package.Execute', 
		          {'uuid': bpformation.args.args.uuid, 'servers': bpformation.args.args.server, 'parameters': bpformation.args.args.parameter })


	def BlueprintList(self):
		self.Exec('bpformation.blueprint.List', {'filters': bpformation.args.args.filter }, cols=['name','id','visibility','date_added'])


	def BlueprintExport(self):
		self.Exec('bpformation.blueprint.Export', {'id': bpformation.args.args.id, 'file': bpformation.args.args.file }, cols=[])


	def BlueprintImport(self):
		self.Exec('bpformation.blueprint.Import', {'files': bpformation.args.args.file }, cols=[])


	def BlueprintUpdate(self):
		self.Exec('bpformation.blueprint.Update', {'files': bpformation.args.args.file }, cols=[])


	def BlueprintDelete(self):
		self.Exec('bpformation.blueprint.Delete', {'ids': bpformation.args.args.id }, cols=[])


	def BlueprintExecute(self):
		self.Exec('bpformation.blueprint.Execute', 
		          {'files': bpformation.args.args.file, 'type': bpformation.args.args.type, 'password': bpformation.args.args.password,
				   'group_id': bpformation.args.args.group_id, 'network': bpformation.args.args.network, 'dns': bpformation.args.args.dns,
				   'parameters': bpformation.args.args.parameter, 'ids': bpformation.args.args.id })


	def Exec(self,function,args=False,cols=None,supress_output=False):
		try:
			if args:  r = eval("%s(**%s)" % (function,args))
			else:  r = eval("%s()" % (function))

			#  Filter results
			if bpformation.args.args.cols:  cols = bpformation.args.args.cols

			# Output results
			# TODO - how do we differentiate blueprints vs. queue RequestIDs?
			#if r is not None and 'RequestID' in r and not bpformation.args.args.async:  
			#	r = bpformation.output.RequestBlueprintProgress(r['RequestID'],self._GetLocation(),self._GetAlias(),bpformation.args.args.quiet)
			#	cols = ['Server']

			if r and len(cols):
				if not isinstance(r, list):  r = [r]
				if not supress_output and bpformation.args.args.format == 'json':  print bpformation.output.Json(r,cols)
				elif not supress_output and bpformation.args.args.format == 'table':  print bpformation.output.Table(r,cols)
				elif not supress_output and bpformation.args.args.format == 'text':  print bpformation.output.Text(r,cols)
				elif not supress_output and bpformation.args.args.format == 'csv':  print bpformation.output.Csv(r,cols)

			#return(r)
		except KeyboardInterrupt:
			bpformation.output.Status("ERROR",3,"Exiting keyboard interupt")
			sys.exit(1)
		except:
			raise


